/* ----------------------------------------------------------------------------
 *         ATMEL Microcontroller Software Support 
 * ----------------------------------------------------------------------------
 * Copyright (c) 2008, Atmel Corporation
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the disclaimer below.
 *
 * Atmel's name may not be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * DISCLAIMER: THIS SOFTWARE IS PROVIDED BY ATMEL "AS IS" AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT ARE
 * DISCLAIMED. IN NO EVENT SHALL ATMEL BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * ----------------------------------------------------------------------------
 */

//-----------------------------------------------------------------------------
/// \dir "Basic EMAC lwIP project"
///
/// !!!Purpose
///
/// This project implements the webserver example of the <a href="../../../external_libs/ethernet/lwip/lwip.pdf"> 
/// lwIP TCP/IP stack</a>. It enables the device to act as a web server, sending a very short page 
/// when accessed through a browser. 
///
/// !!!Hardware Requirements
///
/// The system reset signal NRST is connected to both the Ethernet PHY and the JTAG/ICE interface. 
/// This prevents the project from working properly when a JTAG probe is connected. 
/// To disconnect NRST from the ICE interface, the following modifications must be performed: 
///  - AT91SAM7X/XC-EK: cut strap S2. 
///  - AT91SAM9XE-EK: remove resistor R14 
///  - AT91SAM9260-EK: remove resistor R14 
///  - AT91SAM9263-EK: remove resistor R18 
///
/// !!!See
///
///   - emac: EMAC interface driver
///   - dm9161: PHY interface driver
///
/// !!!Description
///
/// Please refer to the lwIP documentation (included in the package) for more information about 
/// the TCP/IP stack and the webserver example. 
///
/// !!!Usage
///
/// -# Build the program and download it inside the evaluation board. Please
///    refer to the
///    <a href="http://www.atmel.com/dyn/resources/prod_documents/doc6224.pdf">
///    SAM-BA User Guide</a>, the
///    <a href="http://www.atmel.com/dyn/resources/prod_documents/doc6310.pdf">
///    GNU-Based Software Development</a> application note or to the
///    <a href="ftp://ftp.iar.se/WWWfiles/arm/Guides/EWARM_UserGuide.ENU.pdf">
///    IAR EWARM User Guide</a>, depending on your chosen solution.
/// -# On the computer, open and configure a terminal application
///    (e.g. HyperTerminal on Microsoft Windows) with these settings:
///   - 115200 bauds
///   - 8 bits of data
///   - No parity
///   - 1 stop bit
///   - No flow control
/// -# Connect an Ethernet cable between the evaluation board and the network. 
///     The board may be connected directly to a computer; in this case, 
///     make sure to use a cross/twisted wired cable such as the one provided with the evaluation kit. 
/// -# Start the application. It will display the following message on the DBGU: 
///    \code   
///    -- Basic EMAC lwIP Project xxx --
///    -- AT91xxxxxx-xx
///    -- Compiled: xxx xx xxxx xx:xx:xx --
///    - MAC 00:45:56:78:9a:bc
///    - Host IP 10.159.245.156
///    - Gateway IP 10.159.245.157
///    - Net Mask 255.255.254.0
///    \endcode
/// -# Type the IP address (Host IP in the debug log) of the device in a web browser. The page generated by lwIP will appear. 
///
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/// \unit
///
/// !Purpose
///
/// This file contains all the specific code for the
/// basic-emac-lwip-project
///
/// !Contents
///
/// The code can be roughly broken down as follows:
///    - Init EMAC driver structure
///    - Init DM9161 driver
///    - Initialize lwIP modules
///    - Initialize httpd
///    - Run emacif_poll polling tasks
///
/// Please refer to the list of functions in the #Overview# tab of this unit
/// for more detailed information.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
//         Headers
//-----------------------------------------------------------------------------

#include <board.h>
#include <pio/pio.h>
#include <aic/aic.h>
#include <dbgu/dbgu.h>
#include <usart/usart.h>
#include <emac/emac.h>
#include <ethernet/dm9161/dm9161.h>
#include <string.h>
#include <utility/trace.h>
#include <utility/assert.h>

#include "emacif.h"
#include "httpd.h"
#include "arch/sys_arch.h"
#include "tcp.h"
#include "memp.h"
#if LWIP_UDP
#include "udp.h"
#endif
#if LWIP_DHCP
#include "dhcp.h"
#endif

//-----------------------------------------------------------------------------
//         Types
//-----------------------------------------------------------------------------

/// Timer for calling lwIP tmr functions without system
typedef struct _timers_info {
    unsigned int timer;
    unsigned int timer_interval;
    void (*timer_func)(void);
} timers_info;

//-----------------------------------------------------------------------------
//         Local Define
//-----------------------------------------------------------------------------

/// lwIP tmr functions list
static timers_info timers_table[] = {
    /// LWIP_TCP
    { 0, TCP_FAST_INTERVAL,     tcp_fasttmr},
    { 0, TCP_SLOW_INTERVAL,     tcp_slowtmr},
    /// LWIP_ARP
    { 0, ARP_TMR_INTERVAL,      etharp_tmr},
    /// LWIP_DHCP
    #if LWIP_DHCP
    { 0, DHCP_COARSE_TIMER_SECS, dhcp_coarse_tmr},
    { 0, DHCP_FINE_TIMER_MSECS,  dhcp_fine_tmr},
    #endif
};

//-----------------------------------------------------------------------------
//         Local variables
//-----------------------------------------------------------------------------

/// EMAC power control pin
#if !defined(BOARD_EMAC_POWER_ALWAYS_ON)
static const Pin emacPwrDn[] = {BOARD_EMAC_PIN_PWRDN};
#endif

/// The PINs' on PHY reset
static const Pin emacRstPins[] = {BOARD_EMAC_RST_PINS};

/// The PINs for EMAC
static const Pin emacPins[] = {BOARD_EMAC_RUN_PINS};

/// PHY address
#define EMAC_PHY_ADDR             31

/// The DM9161 driver instance
static Dm9161 gDm9161;

/// The MAC address used for demo
static const unsigned char MacAddress[6] = {0x00, 0x45, 0x56, 0x78, 0x9a, 0xbc};

#if !LWIP_DHCP
/// The IP address used for demo (ping ...)
static const unsigned char IpAddress[4] = {10, 159, 245, 156};

/// Set the default router's IP address.
static const unsigned char GateWay[4] = {10, 159, 245, 157};

// The NetMask address
static const unsigned char NetMask[4] = {255, 255, 254, 0};
#endif

//-----------------------------------------------------------------------------
//         Local Functions
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/// Emac interrupt handler
//-----------------------------------------------------------------------------
static void ISR_Emac(void)
{
    EMAC_Handler();
}

//-----------------------------------------------------------------------------
/// Process timing functions
//-----------------------------------------------------------------------------
static void timers_update(void)
{
    static unsigned int last_time;
    unsigned int cur_time, time_diff, idxtimer;
    timers_info * ptmr_inf;
    cur_time = sys_get_ms();
    if (cur_time >= last_time) {
        time_diff = cur_time - last_time;
    }
    else {
        time_diff = 0xFFFFFFFF - last_time + cur_time;
    }
    if (time_diff) {
        last_time = cur_time;
        for(idxtimer=0;
            idxtimer < (sizeof(timers_table)/sizeof(timers_info));
            idxtimer ++) {
            ptmr_inf = &timers_table[idxtimer];
            ptmr_inf->timer += time_diff;
            if (ptmr_inf->timer > ptmr_inf->timer_interval) {
                if (ptmr_inf->timer_func)
                    ptmr_inf->timer_func();
                ptmr_inf->timer -= ptmr_inf->timer_interval;
            }
        }
    }
}

//-----------------------------------------------------------------------------
/// Initialize lwIP modules
//-----------------------------------------------------------------------------
static void lwip_init(void)
{
#if LWIP_STATS
    stats_init();
#endif

    mem_init();
    memp_init();
    pbuf_init();
    etharp_init();
    tcp_init();

#if LWIP_UDP
    udp_init();
#endif
}

//-----------------------------------------------------------------------------
//         Global Functions
//-----------------------------------------------------------------------------

/* The main() function. */
int
main()
{
    struct ip_addr ipaddr, netmask, gw;
    struct netif NetIf, *netif;
#if LWIP_DHCP
    u8_t   dhcp_state = DHCP_INIT;    
#endif

    Dm9161       *pDm = &gDm9161;
    unsigned int errCount = 0;

    TRACE_CONFIGURE(DBGU_STANDARD, 115200, BOARD_MCK);
    printf("-- Basic EMAC lwIP Project %s --\n\r", SOFTPACK_VERSION);
    printf("-- %s\n\r", BOARD_NAME);
    printf("-- Compiled: %s %s --\n\r", __DATE__, __TIME__);

    // Display MAC & IP settings
    printf(" - MAC %x:%x:%x:%x:%x:%x\n\r",
           MacAddress[0], MacAddress[1], MacAddress[2],
           MacAddress[3], MacAddress[4], MacAddress[5]);

#if !LWIP_DHCP
    printf(" - Host IP  %d.%d.%d.%d\n\r",
           IpAddress[0], IpAddress[1],
           IpAddress[2], IpAddress[3]);
    printf(" - Gateway IP  %d.%d.%d.%d\n\r",
           GateWay[0], GateWay[1], GateWay[2], GateWay[3]);
    printf(" - Net Mask  %d.%d.%d.%d\n\r",
           NetMask[0], NetMask[1], NetMask[2], NetMask[3]);
#else
    printf(" - DHCP Enabled\n\r");
#endif

#if !defined(BOARD_EMAC_POWER_ALWAYS_ON)
    // clear PHY power down mode
    PIO_Configure(emacPwrDn, 1);
#endif

    // Init EMAC driver structure
    EMAC_Init(AT91C_ID_EMAC, MacAddress, EMAC_CAF_ENABLE, EMAC_NBC_DISABLE);

    // Init DM9161 driver
    DM9161_Init(pDm, EMAC_PHY_ADDR);

    // PHY initialize
    if (!DM9161_InitPhy(pDm, BOARD_MCK,
                        emacRstPins, PIO_LISTSIZE(emacRstPins),
                        emacPins, PIO_LISTSIZE(emacPins))) {

        printf("P: PHY Initialize ERROR!\n\r");
        return -1;
    }
    // Auto Negotiate
    if (!DM9161_AutoNegotiate(pDm)) {

        printf("P: Auto Negotiate ERROR!\n\r");
        return -1;
    }

    while( DM9161_GetLinkSpeed(pDm, 1) == 0 ) {

        errCount++;
    }
    printf("P: Link detected 0x%X\n\r", errCount);

    // Setup EMAC buffers and interrupts
    AIC_ConfigureIT(AT91C_ID_EMAC, AT91C_AIC_SRCTYPE_INT_HIGH_LEVEL, ISR_Emac);
    AIC_EnableIT(AT91C_ID_EMAC);

    // Initialize system timing
    sys_init_timing();

    // Initialize lwIP modules
    lwip_init();

    // Initialize net interface for lwIP
    emacif_setmac((u8_t*)MacAddress);

#if !LWIP_DHCP
    IP4_ADDR(&gw, GateWay[0], GateWay[1], GateWay[2], GateWay[3]);
    IP4_ADDR(&ipaddr, IpAddress[0], IpAddress[1], IpAddress[2], IpAddress[3]);
    IP4_ADDR(&netmask, NetMask[0], NetMask[1], NetMask[2], NetMask[3]);
#else
    IP4_ADDR(&gw, 0, 0, 0, 0);
    IP4_ADDR(&ipaddr, 0, 0, 0, 0);
    IP4_ADDR(&netmask, 0, 0, 0, 0);
#endif

    netif = netif_add(&NetIf, &ipaddr, &netmask, &gw, NULL, emacif_init, ip_input);
    netif_set_default(netif);
    netif_set_up(netif);

  #if LWIP_DHCP
    dhcp_start(netif);
  #endif

    // Initialize http server application
    if (ERR_OK != httpd_init()) {

        return -1;
    }

    while(1) {

        // Display Statistics
        if ( USART_IsDataAvailable((AT91S_USART *)AT91C_BASE_DBGU) ) {

            EmacStats    stats;

            DBGU_GetChar();
            EMAC_GetStatistics(&stats, 1);
            printf("=== EMAC Statistics ===\n\r");
            printf(" .tx_packets = %d\n\r", stats.tx_packets);
            printf(" .tx_comp = %d\n\r", stats.tx_comp);
            printf(" .tx_errors = %d\n\r", stats.tx_errors);
            printf(" .collisions = %d\n\r", stats.collisions);
            printf(" .tx_exausts = %d\n\r", stats.tx_exausts);
            printf(" .tx_underruns = %d\n\r", stats.tx_underruns);
            printf(" .rx_packets = %d\n\r", stats.rx_packets);
            printf(" .rx_eof = %d\n\r", stats.rx_eof);
            printf(" .rx_ovrs = %d\n\r", stats.rx_ovrs);
            printf(" .rx_bnas = %d\n\r", stats.rx_bnas);
        }

        // Run periodic tasks
        timers_update();

        // Run polling tasks
        emacif_poll(netif);

      #if LWIP_DHCP
        // Dump DHCP addresses
        if (netif->dhcp) {

            u8_t tmp_state = netif->dhcp->state;
            if (tmp_state != dhcp_state) {

                //printf("DHCP: %d -> %d\n\r",
                //    dhcp_state, tmp_state);

                if (tmp_state == DHCP_BOUND &&
                    dhcp_state != DHCP_RENEWING) {

                    u8_t * pAddr;
                    printf("\n\r");
                    printf("=== DHCP Configurations ===\n\r");
                    pAddr = (u8_t*)&netif->ip_addr;
                    printf("- IP   : %d.%d.%d.%d\n\r",
                        pAddr[0], pAddr[1], pAddr[2], pAddr[3]);
                    pAddr = (u8_t*)&netif->netmask;
                    printf("- Mask : %d.%d.%d.%d\n\r",
                        pAddr[0], pAddr[1], pAddr[2], pAddr[3]);
                    pAddr = (u8_t*)&netif->gw;
                    printf("- GW   : %d.%d.%d.%d\n\r",
                        pAddr[0], pAddr[1], pAddr[2], pAddr[3]);
                    printf("===========================\n\r\n");
                }
                dhcp_state = tmp_state;
            }
        }
      #endif
    }
    return 0;
}
